/****************************************************************************
*
*   Copyright (c) 2005 Dave Hylands
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
****************************************************************************/

/* Flash-LED.c modified by Dima Diall to read ADC on-demand */

#include <avr/io.h>
#include <stdio.h>

#include "a2d.h"
#include "Hardware.h"
#include "Delay.h"
#include "UART.h"


  /*
  \x02 // STX - start of text
  \x08 // Number of Bytes in Core Message
  \xAC // Checksum - Sum of decimal byte values Modules 256

  // --- START Core Message ---

  \x30 // Board address Byte 2 i.e. Most Significant Byte
  \x31 // Board address Byte 1 i.e. Least Significant Byte

  \x35 // Command 5

  // LEDs On(0)/Off(1) - B=Blue, G=Green, Y=Yellow, R=Red, X=eXtra(not used)
  \x00 // XXXX,XXXX
  \x00 // XXXX,XXRR
  \x00 // XXXX,XXXX
  \x00 // RYYY,GGGB

  \x57 //w - write operation

  // --- END Core Message ---

  \x03 // ETX - end of text
  */


/**
 * Utility function; useful for debugging i.e. inspecting byte value
 *
 * Displays a byte(char i.e. 8 bits) on the LEDs. Each bit
 * is represented by an LED. A bit with the value 1 will have
 * its corresponding LED switched On. A bit with the value 0 
 * will have its corresponding LED switched Off.
 *
 * The single Blue LED is the least significant bit of the byte to be displayed
 * The first Red LED(starting from the blue LED side) is the most significant
 * bit.
 * The two remaining red LEDs are ignored as a byte only needs 8 LEDS to
 * represent its 8 bits.
 */
int byteToLeds(const char byte)
{
  unsigned short int MSG_SIZE = 12;
  unsigned short int MSG_HEADER_SIZE = 3;
  unsigned short int MSG_TRAILER_SIZE = 1;

			 // 0,	1,	 2,	 3,4,	 5,	 	6,7,8,9,	 10,	11
			 // STX,SIZE,SUM,ADDRESS,CMD,LED,LED,LED,LED,WRITE,ETX
  char msg[] = "\x02\x08\xEE\x30\x30\x35\x00\xFF\x00\xFF\x57\x03"; // valid sample output
  int i;

	// Sets all LEDs to OFF status
	//msg[6] = 'F'; msg[7] = 'F'; msg[8] = 'F'; msg[9] = 'F';
/*
  // LEDs On(0)/Off(1) - B=Blue, G=Green, Y=Yellow, R=Red, X=eXtra(not used)
  \x00 // XXXX,XXXX // char 6
  \x00 // XXXX,XXRR // char 7
  \x00 // XXXX,XXXX // char 8
  \x00 // RYYY,GGGB // char 9
*/
	// LEDs On(0)/Off(1)
	// example:
	// value of msg[9] is 0xFF
	// if bit zero is '1' then XOR; 0xFF ^ 0x01 and store the results in msg[9]
	// so msg[9] now holds 0xFE
	// Therefore every time a bit is set the appropriate bit in the message gets set to zero
	// ... which switches the appropriate LED on.
	// we could instead set the intial value to 0x00 and then use bitwise or operator
	// and at the end use the bitwise not operator to reverse bit pattern,
	// that would probably make more sense, i.e. be easier to digest what is actually happening
	if( (byte & 0x01) == 0x01 ) msg[9] ^= 0x01;
	if( (byte & 0x02) == 0x02 ) msg[9] ^= 0x02;
	if( (byte & 0x04) == 0x04 ) msg[9] ^= 0x04;
	if( (byte & 0x08) == 0x08 ) msg[9] ^= 0x08;

	if( (byte & 0x10) == 0x10 ) msg[9] ^= 0x10;
	if( (byte & 0x20) == 0x20 ) msg[9] ^= 0x20;
	if( (byte & 0x40) == 0x40 ) msg[9] ^= 0x40;
	if( (byte & 0x80) == 0x80 ) msg[9] ^= 0x80;

	// message sum (checksum)
	msg[2] = '\x00';
	// Calculates(Adds) the message sum.
	for( i = 3; i < MSG_SIZE-1; i++) msg[2] += msg[i];

	// calculates message size
	msg[1] = MSG_SIZE - MSG_HEADER_SIZE - MSG_TRAILER_SIZE;
	for( i = 0; i < MSG_SIZE; i++) UART1_PutChar(msg[i]);

	return 0;
}


/**
 *
 *
 * Test function; lights up all leds on all boards 1 by 1 with 0.2 seconds delay
 **/
int writeToPort( FILE *u1 )
{
  unsigned short int MSG_SIZE = 12;
  unsigned short int MSG_HEADER_SIZE = 3;
  unsigned short int MSG_TRAILER_SIZE = 1;

// valid sample output with all the LEDs in ON status!!!
			 // 0,	1,	 2,	 3,4,	 5,	 6,7,8,9,		 10,	11
			 // STX,SIZE,SUM,ADDRESS,CMD,LED,LED,LED,LED,WRITE,ETX
  char msg[] = "\x02\x08\xEF\x30\x32\x35\x00\xFF\x00\xFF\x57\x03";
  int i;
  int leds;
  int boards;

	//InitTimer();
	msg[4] = '\x2F';
	for(boards = 0; boards < 4; boards++)
	{
		msg[7] = '\xFF';
		msg[9] = '\xFF';
		msg[4]++;
	for (leds = 0; leds < 11; leds++)
	{
		// message sum (checksum)
		msg[2] = '\x00';
		// Calculates(Adds) the message sum.
		for( i = 3; i < MSG_SIZE-1; i++)
		{
			// adds to message sum
	  		msg[2] += msg[i];
		}
		// calculates message size
	      	msg[1] = MSG_SIZE - MSG_HEADER_SIZE - MSG_TRAILER_SIZE;

		for( i = 0; i < MSG_SIZE; i++)
		{
			UART1_PutChar(msg[i]);
		}

		// Tick rate is 100/sec so waiting for 100 waits for 1 sec
		for ( i = 0; i < 10; i++ )
		{
			WaitForTimer0Rollover();
		}

		if( leds < 8) msg[9] <<= 1;
		else msg[7] <<= 1;
    }// end for loop, switching on leds 1 each second
	}// end loop going through each led board

	//fprintf(u1, "\x02\x08\xE9\x30\x31\x35\x00\xFF\x00\xFD\x57\x03");
  return 0;
}

/************************************************************************
 *
 * displays sensor reading on a spectrum of 1-10
 *
 **/
int displaySensorReading(char channel, uint16_t reading)
{
  unsigned short int MSG_SIZE = 12;
  unsigned short int MSG_HEADER_SIZE = 3;
  unsigned short int MSG_TRAILER_SIZE = 1;

  long int current = 1;
  long int increment;
  // Default values given to max and min in case
  // fake results needs to be generated.
  long int max = 0x0391;
  long int min = 0x0110;

  char ledBoardAddress = '\x30';

			 // 0,	1,	 2,	 3,4,	 5,	 6,7,8,9,		 10,	11
			 // STX,SIZE,SUM,ADDRESS,CMD,LED,LED,LED,LED,WRITE,ETX
  char msg[] = "\x02\x08\xEF\x30\x32\x35\x00\xFF\x00\xFF\x57\x03";//valid sample output which places all the LEDs to OFF
  int i;


  switch(channel)
  {
  	case '0':								// Microphone i.e. acoustic sensor
		ledBoardAddress = '\x30';
		max = 0x0391;
		min = 0x0120;
		increment = ( (max - min) / 10);
		current = (reading - min) / increment;
		break;
  	case '1':								// CO2 Sensor
		ledBoardAddress = '\x31';
		max = 0x0155;
		min = 0x0100;
		increment = ( (max - min) / 10);
		current = (reading - min) / increment;
        break;
	case '2':								// CO Sensor
		ledBoardAddress = '\x32';
		max = 0x00b5;
		min = 0x0040;
		increment = ( (max - min) / 10);
		current = (reading - min) / increment;
        break;
    case '3':								// CSV - Solvent Vapor Sensor (e.g. Alcohol)
    	ledBoardAddress = '\x33';
		max = 0x01c9;
		min = 0x011a;
		increment = ( (max - min) / 10);
		current = (reading - min) / increment;
        break;
    // we only have 4 led boards so...
    case '4':								// Battery voltage
    	return 0;
    	break;
    case '5':
	case '6':
	case '7':
	default:
    	//ledBoardAddress = '\x30';
    	current = 1;
    	return 0;
		break;
	}

	// assigns LED board address
	//msg[3] = '\x30';
	msg[4] = ledBoardAddress;
/*
  // LEDs On(0)/Off(1) - B=Blue, G=Green, Y=Yellow, R=Red, X=eXtra(not used)
  \x00 // XXXX,XXXX // char 6
  \x00 // XXXX,XXRR // char 7
  \x00 // XXXX,XXXX // char 8
  \x00 // RYYY,GGGB // char 9
*/
		if(current >= 10)		{ msg[7] = '\x03'; msg[9] = '\xFF'; }
		else if( current >= 9 ) { msg[7] = '\x01'; msg[9] = '\xFF'; }
		else if( current >= 8 ) { msg[7] = '\x00'; msg[9] = '\xFF'; }
		else if( current >= 7 ) { msg[7] = '\x00'; msg[9] = '\x7F'; }
		else if( current >= 6 ) { msg[7] = '\x00'; msg[9] = '\x3F'; }
		else if( current >= 5 ) { msg[7] = '\x00'; msg[9] = '\x1F'; }
		else if( current >= 4 ) { msg[7] = '\x00'; msg[9] = '\x0F'; }
		else if( current >= 3 ) { msg[7] = '\x00'; msg[9] = '\x07'; }
		else if( current >= 2 )	{ msg[7] = '\x00'; msg[9] = '\x03'; }
		//else if( current >= 1 ) { msg[7] = '\x00'; msg[9] = '\x01'; }
		else					{ msg[7] = '\x00'; msg[9] = '\x01'; }

		msg[9] = ~ msg[9];
		msg[7] = ~ msg[7];
		

	// message sum (checksum)
	msg[2] = '\x00';
	// Calculates(Adds) the message sum.
	for( i = 3; i < MSG_SIZE-1; i++)
	{
		// adds to message sum
		msg[2] += msg[i];
	}
	// calculates message size
	msg[1] = MSG_SIZE - MSG_HEADER_SIZE - MSG_TRAILER_SIZE;
	for( i = 0; i < MSG_SIZE; i++)
	{
		UART1_PutChar(msg[i]);
	}

	return 0;
}// end displaySensorReading( uint16_t reading)



int main(void)
{
    FILE	*u0;
    FILE	*u1;
    char channel;
    uint16_t reading = 0;

    InitHardware();
    InitTimer();

    // The first handle opened for read goes to stdin, and the first handle
    // opened for write goes to stdout. So u0 is stdin, stdout, and stderr
#if defined( __AVR_LIBC_VERSION__ )
    u0 = fdevopen( UART0_PutCharStdio, UART0_GetCharStdio );
    u1 = fdevopen( UART1_PutCharStdio, UART1_GetCharStdio );
#else
    u0 = fdevopen( UART0_PutCharStdio, UART0_GetCharStdio, 0 );
    u1 = fdevopen( UART1_PutCharStdio, UART1_GetCharStdio, 0 );
#endif


	//LED_ON(USB_RED);	//USB stick in use
//	LED_ON(USB_GREEN);	// USB stick Not in use
	//LED_ON(SENSOR_PREHEAT);		// Sensor pre-heat period
	//LED_ON(GPS);					//GPS Fix

	DEVICE_ON( SENSOR_BOARD );
	DEVICE_ON( LED_BOARD );

	// switches all LEDs On one by one, using a small pause between each
	writeToPort(u1);
	
	// Indicates the system is ready - First LED on the box (Green)
	LED_ON( SYS );		// System ready LED

    while( 1 )	// infinite loop
    {
        if ( UART0_IsCharAvailable() ) // Polls UART for incoming characters
        {

            // Gets ADC channel number as ASCII character
			channel = getchar() - '0';

            // Turns off all of the robostix' onboard LEDs
            LED_OFF( RED );		LED_OFF( BLUE );	LED_OFF( YELLOW );

	    	// Verify if ADC channel is in range (0..7)
            if ( channel >= 0 && channel <= 7 )
            {
                LED_ON( BLUE );

				// On ADC channel 2 aka F.2 the CO sensor is connected.
				// CO is located at F.2
				if(channel == 2)
				{
					DEVICE_ON(CO_HEATER);
					Delay1mSec(14);
					DEVICE_OFF(CO_HEATER);
					Delay1mSec((uint8_t)986);
					DEVICE_ON(CO_SENSOR);
					Delay100uSec(25); //Delay1mSec(2.5);
					reading = a2d_10( channel );
					Delay100uSec(25); //Delay1mSec(2.5);
					DEVICE_OFF(CO_SENSOR);
					Delay1mSec((uint8_t)995); // ensures cool down period for CO sensor
				}
				else
				{
					// Read requested ADC channel
					reading = a2d_10( channel );
				}
				
				// prints result (in hex) to the serial port which connects to gumstix
				printf( "0x%04x\n", reading /* a2d_10( channel ) */);

				// sends readings to LED board for real time in-situ display of sensor readings
				displaySensorReading(channel+'0', reading);

				//IF channel is F.2 (i.e. CO) then
                // turn on pin C.3 for X milliseconds (Sensor On) then
                // turn on pin C.2 for X millliseconds (Heater On) then
                // take reading, i.e. a2d_10(ch)
                //ELSEIF
                // certain non-existing channel/ char is sent to indicate
                // GPS status appropriate LED On/Off
                //ELSEIF (covering cases for each system LED, 
				//			however this is to be moved out of the outer if statement.
				//			and rather proceed it?
                // ...
                // ...
                // ...
                //ELSE
                // take reading and print to serial as usual
                //END
                
                // Is it worth switching the whole sensor board On/Off between each reading ?
                // to conserve energy (Port C.1)
            }
            else
            {
                // Channel out of range: error!
                LED_ON( RED );
                printf( "???\n" );
            }
        }
        // Do other work, introduce some delay, go to sleep (...)
    }//end while

    return 0;
}

