/****************************************************************************
FERAL ROBOT - URBAN TAPESTRIES (http://urbantapestries.org/)
Linux-based robot client for gumstix platform
Author: Dima Diall
****************************************************************************/

#include "sensordef.h"
#include <stdio.h>
#include <stdlib.h>

// Function prototypes for sensor value conversion handlers
float conv_inpvolt(uint16_t adc_val);
float conv_battery(uint16_t adc_val);
float conv_air_co2(uint16_t adc_val);
float conv_airqual(uint16_t adc_val);

int proc_inpvolt(float input_voltage);
int proc_battery(float battery_level);

// Sensor definition table, one entry for each ADC channel or sensor
const sensor_t sensor_defs[ADC_CHANNELS] = {
{ 1, "INPV",    conv_inpvolt,  proc_inpvolt,   0 },
{ 1, "BATT",    conv_battery,  proc_battery,   0 },
{ 1, "CO_2",    conv_air_co2,  NULL,           1 },
{ 1, "AIRQ",    conv_airqual,  NULL,           1 },
{ 0, NULL,      NULL,          NULL,           0 },
{ 0, NULL,      NULL,          NULL,           0 },
{ 0, NULL,      NULL,          NULL,           0 },
{ 0, NULL,      NULL,          NULL,           0 }
};

// Potentiometer divisor settings for ADC channel input
#define INPVOLT_DIV         1.0
#define BATTERY_DIV         2.0
#define AIR_CO2_DIV         1.0
#define AIRQUAL_DIV         1.0

// Maximum real sensor value
#define INPVOLT_MAX         5.0
#define BATTERY_MAX         (1.4 * 7)
#define AIR_CO2_MAX         3000.0
#define AIRQUAL_MAX         2.5

// Minimum real sensor value
#define INPVOLT_MIN         3.4
#define BATTERY_MIN         (1.2 * 7)
#define AIR_CO2_MIN         0.0
#define AIRQUAL_MIN         0.75

// Convert input voltage ADC reading to real value.
// Input range:  up to 5.0V (divided by INPVOLT_DIV)
// Output range: up to 5.0V
float conv_inpvolt(uint16_t adc_val)
{
    float result = CONV_ADC(adc_val) * INPVOLT_DIV;
    return result;          /* output in volts */
}

// Convert battery voltage ADC reading to real value.
// Input range:  up to 9.8V (divided by BATTERY_DIV)
// Output range: up to 9.8V
float conv_battery(uint16_t adc_val)
{
    float result = CONV_ADC(adc_val) * BATTERY_DIV;
    return result;          /* output in volts */
}

// Convert CO_2 sensor ADC reading to real value.
// Input range:  0.0 - 3.0V (divided by AIR_CO2_DIV)
// Output range: 0 - 3000ppm
float conv_air_co2(uint16_t adc_val)
{
    float result = CONV_ADC(adc_val) * AIR_CO2_DIV;
    return result * 1000.0;     /* convert to ppm units */
}

// Convert air quality sensor ADC reading to real value.
// Input range:  0.7 - 2.5V (divided by AIRQUAL_DIV)
// Output range: ???
float conv_airqual(uint16_t adc_val)
{
    float result = CONV_ADC(adc_val) * AIRQUAL_DIV;
    return result;
}


// Handler for input voltage level.
// Returns 0 if input voltage outside normal range.
int proc_inpvolt(float input_voltage)
{
    return (input_voltage < INPVOLT_MIN || input_voltage > INPVOLT_MAX);
}

// Handler for battery charge level.
// Terminates program if battery voltage drops below a certain threshold.
int proc_battery(float battery_level)
{
    if(battery_level < BATTERY_MIN) {
        fprintf(stderr, "\nBattery level critical (%02fV, min: %02fV)!\n\n",
                battery_level, BATTERY_MIN);
        // we could terminate with exit code 100
        // to signal system shutdown to wrapper script
        //exit(100);    // terminate program execution

        return 100;
    }
    
    return 0;
}
