/****************************************************************************
FERAL ROBOT - URBAN TAPESTRIES (http://urbantapestries.org/)
Linux-based robot client for gumstix platform
Author: Dima Diall
****************************************************************************/

#include "sensordef.h"
#include <stdio.h>
#include <stdlib.h>

// Function prototypes for sensor value conversion handlers
float conv_microphone(uint16_t adc_val);
float conv_air_co2(uint16_t adc_val);
float conv_air_co(uint16_t adc_val);
float conv_air_csv(uint16_t adc_val);
float conv_battery(uint16_t adc_val);

int proc_battery(float battery_level);

// Sensor definition table, one entry for each ADC channel or sensor
// Sensor definition entry, which corresponds to an ADC channel.
//     is_active       whether ADC input is connected to a sensor
//     sensor_id       string describing database sensor id (4 chars)
//     adc_handler     pointer to ADC output conversion function
//     value_handler   pointer to sensor value processing handlers
//     transmit        whether to transmit sensor value to UT server
const sensor_t sensor_defs[ADC_CHANNELS] = {
//up,	name,	conversion_function, 		???, 	Send to Server	// ADC Channel(i.e. port/pin)
{ 1, "MICR",	conv_microphone,			NULL,           0 },	// F.0
{ 1, "CO_2",	conv_air_co2,				NULL,           0 },	// F.1
{ 1, "CO__",	conv_air_co,				NULL,           0 },	// F.2
{ 1, "CSV_",	conv_air_csv,				NULL,           0 },	// F.3
{ 1, "BATT",	conv_battery,				proc_battery,   0 },	// F.4
{ 0, NULL,		NULL,          				NULL,           0 },
{ 0, NULL,      NULL,          				NULL,           0 },
{ 0, NULL,      NULL,          				NULL,           0 }
};

// Potentiometer divisor settings for ADC channel input
// ***
#define INPVOLT_DIV         1.0
#define BATTERY_DIV         4.0			// one fourth(1/4) of the battery voltage level is supplied to F.4
#define AIR_CO2_DIV         1.0
//#define AIRQUAL_DIV         1.0

// *** Maximum real sensor value
#define INPVOLT_MAX         5.0
#define BATTERY_MAX         (1.4 * 7)
#define AIR_CO2_MAX         3000.0
#define AIRQUAL_MAX         2.5

// Minimum real sensor value
// *** 
#define INPVOLT_MIN         3.4
// TODO: 7? 8? batteries
// FIXME:
#define BATTERY_MIN         (1.2 * 7)
// ***
#define AIR_CO2_MIN         0.0
// ***
#define AIRQUAL_MIN         0.75



float conv_microphone(uint16_t adc_val)
{
	return (adc_val - 0x0120) / ( (0x0391 - 0x0120) / 10 );
}



float conv_air_co2(uint16_t adc_val)
{
	/* OLD Code
	// Convert CO_2 sensor ADC reading to real value.
	// Input range:  0.0 - 5.0V (divided by AIR_CO2_DIV)
	// Output range: 0 - 3000ppm
    float result = CONV_ADC(adc_val) * AIR_CO2_DIV;
    return result * 1000.0;     //convert to ppm units
    */
	return (adc_val - 0x0100) / ( (0x0155 - 0x0100) / 10 );
}


float conv_air_co(uint16_t adc_val)
{
 	return
	(
	(adc_val - 0x0000) / ( (0x000b - 0x0001) / 10 ) < 0	?
	0.0													:
	(adc_val - 0x0000) / ( (0x000b - 0x0001) / 10 )
	);
}



float conv_air_csv(uint16_t adc_val)
{
	return (adc_val - 0x011a) / ( (0x0266 - 0x011a) / 10 );
}


// Convert battery voltage ADC reading to real value.
// Input range:  up to 9.8V (divided by BATTERY_DIV)
// Output range: up to 9.8V
float conv_battery(uint16_t adc_val)
{
    float result = CONV_ADC(adc_val) * BATTERY_DIV;
    return result;          /* output in volts */
}


// Handler for battery charge level.
// Terminates program if battery voltage drops below a certain threshold.
int proc_battery(float battery_level)
{
    if(battery_level < BATTERY_MIN)
	{
        fprintf(stderr, "\nBattery level critical (%02fV, min: %02fV)!\n\n",
                battery_level, BATTERY_MIN);
        // we could terminate with exit code 100
        // to signal system shutdown to wrapper script
        //exit(100);    // terminate program execution

        return 100;
    }
    return 0;
}
