/****************************************************************************
  FERAL ROBOT - URBAN TAPESTRIES (http://urbantapestries.org/)
  Linux-based robot client for gumstix platform
  Author: Dima Diall
****************************************************************************/

//#include <stdio.h>   /* Standard input/output definitions */
//#include <string.h>  /* String function definitions */
//#include <unistd.h>  /* UNIX standard function definitions */
//#include <fcntl.h>   /* File control definitions */
//#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */

#include "config.h"
#include "utrobot.h"
#include "logger.h"
#include "adchw.h"
#include "sensordef.h"
#include <gps.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>

// Output streams for logging normal data, errors and debugging info
#define DAT         stdout
#define DBG         stderr
#define ERR         stderr

// X = online?, S = fix status?, M = fix mode, P = position, W = watcher [see gpsd(8)]
#define GPS_INIT_STRING    "XW+"

/**
 *
 **/
static void update_position(struct gps_data_t *gpsdata,
                            char *message, size_t len, int level)
{
    log_msg(DBG, LL_TRACE, "GPS update: stat=%u, mode=%u, latitude=%.2f, longitude=%.2f\nmessage='%s'\n",
            gpsdata->status, gpsdata->fix.mode,
            gpsdata->fix.latitude, gpsdata->fix.longitude,
            message);

     /*
     // update global variables
     gps_fix = (gpsdata->fix.mode >= MODE_2D);
     if(gps_fix) {
         gps_latitude = gpsdata->fix.latitude;
         gps_longitude = gpsdata->fix.longitude;
     }
     */
}

// TODO: provide real operating system retrieval of MAC address
// Dummy function to return MAC address string (...)
int get_mac_address(unsigned long long addr_bin, char *mac_addr, int len)
{
    uint16_t hi = htons((addr_bin >> 32) & 0xFFFF);
    uint32_t lo = htonl(addr_bin & 0xFFFFFFFF);
    char *hi_p = (char *) &hi;
    char *lo_p = (char *) &lo;
    int i;

    if(mac_addr == NULL || len < LEN_MAC)
        return -1;

    mac_addr[0] = hi_p[0];
    mac_addr[1] = hi_p[1];
    mac_addr[2] = lo_p[0];
    mac_addr[3] = lo_p[1];
    mac_addr[4] = lo_p[2];
    mac_addr[5] = lo_p[3];

    return 0;
}


char *cut_server_port(char *server_port)
{
    char *colon = strchr(server_port, ':');

    if (colon != NULL) {
        // cut original string at ':' and return pointer to substring after colon
        *colon = '\0';
        return colon + 1;
    }
    return NULL;
}

#define GETOPT_STRING       "i:u:g:a:lvdVh"
/**
 * Prints the command line options for the application program to stderr,
 * and exits the application(with 1)
 *
 **/
void usage(char *progname)
{
    fprintf(stderr,
            "Usage: %s -i hwaddr -u addr[:port] [-g addr[:port]] [-a serial_port] [-ldvh]\n\n"
            "  -i hwaddr        Robot id as MAC address (6 bytes in hex)\n"
            "  -g addr[:port]   GPSD server (default: " GPSD_ADDR ":" GPSD_PORT ")\n"
            "  -a serial_port   Serial port for ADC access (default: " ADC_SERIAL_PORT ")\n"
            "  -l               Log sensor data to standard output (stdout)\n"
            "  -v               Print extra information (on stderr)\n"
            "  -d               Print debugging information (on stderr)\n"
            "  -V               Show version, then exit\n"
            "  -h               Show this help message, then exit\n\n"
            "More information on Feral Robots at http://socialtapestries.net/feralrobots/\n"
            , progname
			);

    exit(1);
}




/**********************************************************************
 *
 * Closes the port which was opened using the argument file descriptor
 * @parameter	fd		file descriptor with value of open port
 *
 */
int closePort(int fd)
{
  close(fd);
  return 0;
}// end closePort(int fileDescriptor)


int setDefaultSettings(int fd, struct termios *settings)
{
  tcsetattr(fd,TCSANOW,settings);
  return 0;
}


/******************************************************************
 *
 * Reads data and dumps it on the screed. Uses the file descriptor parameter.
 *
 */
int readFromPort(int fd)
{
  /* fcntl(fd, F_SETFL, 0); */

  char buffer[255];  // Input character buffer
  char *bufptr;      /* Current char in buffer */
  int  nBytes;       /* Number of bytes read */

  nBytes = read(fd,buffer,255);
  if(nBytes < 0)
    {
      return -1;
    }
  buffer[nBytes] = 0;             /* set end of string, so we can printf */
  //    buffer[nBytes] = '\0';
  printf("%s[%d]\n", buffer, nBytes);

  /* prints buffer contents(characters received from serial port)*/
  //printf("%s", buffer);
  return 0;
} // end readFromPort(int fileDescriptor)




int writeToPort(int fd)
{
  int n = 0;
  /* write function returns number of bytes sent or -1 if an error occurred */
  
  // the following command fails, but that's not a problem because
  // the current device is 1 anyway. i.e. device 2 which is the usb storage
  //readFromPort(fd);
  n = write(fd, "SC 1\r", 5);	/* Sets current device to 1(port B) */
  printf("SC 1\n");
  readFromPort(fd);
  sleep(1);

  /* the following statements are not necessary because they are the default
   * values for the device anyway, so unless they have been changed, 
   * they shouldn't have any effect at all. But we will leave them here,
   * just in case someone has changed the settings! */
  //n = write(fd, "fbd $384100\r", 12);	/* Sets baud rate to 9600 */
  //sleep(1);
  //n = write(fd, "fbd $1A0000\r", 12);
  if (n < 0) { fputs("write() failed!\n", stderr); }
  
  /* Sets RTS(Request To Send) and DTR(Data Terminal Ready) active */
  //[0][0][0][0] [0][0][RTS][DTR], [0][0][0][0] [0][0][Change RTS][Change DTR]
  //n = write(fd, "fmc $0303\r", 10);
  //if (n < 0) { fputs("write() failed!\n", stderr); }

  /* Sets RTS / CTS flow control */
  //[0][0][0][0] [0][XOFF][DTR/DSR][RTS/CTS]
  //n = write(fd, "ffc $01\r", 8);

  //n = write(fd, "QP2\r", 4);
  		sleep(1);
      	n = write(fd, "ECS\r", 4);	// Switches device to extended command set
      	sleep(1);
		n = write(fd, "IPA\r", 4);	// Sets device to input numbers in ASCII
		sleep(1);

  printf("before write loop\n");
  int i = 0;
  while(1)
    {
		i++;
		n = write(fd, "OPW II.TXT\r", 11); //Opens/Creates file for writing
		//      readFromPort(fd);
		if( n < 0) continue;
		sleep(1);
		n = write(fd, "WRF 2\r", 6);	// Tells device to write next X bytes to file
		sleep(1);
		n = write(fd, "hi", 2);			// Writes characters to file
		//    readFromPort(fd);
		if( n < 0) continue;
		sleep(1);
		n = write(fd, "CLF II.TXT\r", 11);//closes file
		//      readFromPort(fd);
		if( n < 0) continue;
		printf("Try: %d \n", i);
		if (i > 0) break; // 3 tries before loop ends.
    }

  if (n < 0)
    {
      fputs("ERROR: write() failed!\n", stderr);
    }
  else
    {
      printf("INFO: Wrote [%d] bytes.\n", n);
    }
  return n;
} // end writeToPort(int fileDescriptor)


/**********************************************************************
 *
 *
 **/
int applyNewSettings(int fd, struct termios *settings)
{
	if ( tcgetattr(fd, settings) )       	// Gets current settings for serial port
    {
      printf("ERROR: Getting termios settings.\n");
    }
	cfsetispeed(settings, B9600);		// Sets input baud rate
	cfsetospeed(settings, B9600);		// Sets output baud rate

	// Control options
	(*settings).c_cflag &= ~CSIZE;		//zero out bits using the character-size mask
	(*settings).c_cflag |= CS8;			// 8bit data, no parity, 1 stop bit
	(*settings).c_cflag &= ~PARENB;		// no parity
	(*settings).c_cflag &= ~(CSTOPB);	// not 2 stop bits (i.e. only 1 stop bit)
	(*settings).c_cflag &= ~CRTSCTS;    // disables Hardware RTS/CTS
  //(*settings).c_cflag |= CRTSCTS;		// disables Hardware RTS/CTS
	(*settings).c_cflag |= (CLOCAL
  			  | CREAD
  			  );
  // Local options
  (*settings).c_lflag |= (ICANON        // Canonical input
								//
			  //| ECHOK    	// Echoes \n after kill character

  			  );
	(*settings).c_lflag &= ~(IEXTEN | ISIG | ECHO | ECHOCTL | ECHOE | ECHOK | ECHOKE | ECHOPRT);

	//Input filtering options (received data from remote device)
	(*settings).c_iflag &= ~(IXON | IXOFF | IXANY);//turns off software flow-control
	//(*settings).c_iflag |= (IXON | IXOFF | IXANY);
	(*settings).c_iflag |= (ICRNL); //Maps \r to \n on the received data
	(*settings).c_iflag &= ~(IGNBRK);		//


   // Output options (data sent from this local device to the remote device)
  //(*settings).c_oflag &= ~OPOST;         // Raw output () i.e. disables processing of output characters
	(*settings).c_oflag |= (OPOST | ONLCR);

	if( tcsetattr(fd, TCSANOW, settings) )   //Sets new options for port immediately
    {
      printf("ERROR: Applying new termios settings.\n");
    }
	else
	{
		printf("INFO: Applied new termios settings.\n");
	}
  return 0;
}//end applySettings(int fileDescriptor, * termios)



int getDefaultSettings(int fd, struct termios *settings)
{
  tcgetattr(fd, settings);
  return 0;
}


/******************************************************************
 *
 * Returns the file descriptor on success or -1 on error.
 * TODO: Pass port id as argument
 *
 **/
int openPort(void)
{
  int fd;                // File Descriptor for the serial port

  fd = open("/dev/ttyS0",  // port name
	    O_RDWR 	   // open for both reading and writing
	    | O_NOCTTY 	   // not the controlling terminal
	    //  	    | O_NDELAY     // not care for state of DCD(Data Carrier Detect)
	    );

  if (fd == -1)		// Could not open the port.
    {
      perror("ERROR: opening port failed. \n");
    }
  else
    {
      printf("INFO: Port opened successfully.\n");
    }
  return (fd);
} // end openPort()



int initializeDevice(int fd)
{
  if( readFromPort(fd) < 0      // BLANK LINE - [\r]
      || readFromPort(fd) < 0   // Version - [Ver 02.19VDAPB On-Line:\r]
      || readFromPort(fd) < 0   // Device detected port - [Device Detected P2\r]
      || readFromPort(fd) < 0   // Found firmware upgrade image file - [Found It\r]
      || readFromPort(fd) < 0   // upgrade firmware if newer version - [No Upgrade\r]
      || readFromPort(fd) < 0 ) // device identifier - [D:\>\r]
    {
      return -1;
    }
      return 0;
}



/****************************************************
 *
 *
 *
 **/
int usb_flash_main()
{
  int fd;
  struct termios backupSettings;
  struct termios settings;

  fd = openPort();
  //getDefaultSettings(fd, &backupSettings);
  applyNewSettings(fd, &settings);

 // if( initializeDevice(fd) < 0 )
   // {
     // printf("ERROR: could not read from port.\n");
    //}
  //else
    //{

      writeToPort(fd);
   //}
  //setDefaultSettings(fd, &backupSettings);
  closePort(fd);
  return 0;
} // end main()



/**
 * Process the command line arguments,
 *
 *
 *
 **/
int main(int argc, char **argv)
{
    int option, i = 0;
    int sock_fd, adc_fd;
    struct sockaddr_in dst_addr;
    struct hostent *hostinfo;
    struct gps_data_t *gpsdata;
    unsigned long long robot_id = 0x0;
    char *utsrv_addr = NULL;
    char *utsrv_port = NULL;
    unsigned utsrv_port_num = UT_SERVER_PORT;
    char *gpsd_addr = GPSD_ADDR;
    char *gpsd_port = GPSD_PORT;
    char *adc_serial = ADC_SERIAL_PORT;
    char ip_str[INET_ADDRSTRLEN];
    uint16_t adc_val;
    int sensor_proc;
    packet_t packet;

    // initialize logger with minimal output (errors and warnings)
    logger_level = LL_FATAL | LL_ERROR | LL_WARN ; //| LL_INFO | LL_DEBUG | LL_TRACE;

/*jt	printf("usb-main starts... \n");
	usb_flash_main();
	printf("Ending usb-main()\n");
	return 0; 
*/

    // process command line arguments
    //printf("Processing command line options.\n");
    while( (option = getopt(argc, argv, GETOPT_STRING)) != -1 )
	{
        switch(option)
		{
            case 'g':
                gpsd_addr = optarg;
                if(gpsd_port = cut_server_port(gpsd_addr))
				{
                    if(strlen(gpsd_addr) < 1 || strlen(gpsd_port) < 1)
                        gpsd_addr = gpsd_port = NULL;
                }
				else
				{
                    gpsd_port = GPSD_PORT;
                }
				log_msg(ERR, LL_FATAL, "gpsd_port: %s - gpsd_addr: %s\n", gpsd_port, gpsd_addr);
                break;

            case 'i': robot_id   = strtoull(optarg, NULL, 16);  break;
            case 'a': adc_serial = optarg;                      break;

            case 'l': logger_level |= LL_INFO;                  break;
            case 'v': logger_level |= LL_DEBUG;                 break;
            case 'd': logger_level |= LL_DEBUG | LL_TRACE;      break;

            case 'V':
                fprintf(stderr, PACKAGE " " VERSION "\n");
                exit(0);
                break;

            case 'h':
            default:
                usage(argv[0]);
                break;
        }//end switch
    }//end while


    log_msg(DBG, LL_TRACE,
            "Finished processing command-line options:\n"
            "Logger level:     %02x\n"
            "Robot id:         %08llx\n"
            "GPSD server:      %s:%s\n"
            "ADC serial port:  %s\n\n",
            logger_level, robot_id,
			gpsd_addr, gpsd_port, adc_serial);

    // verify if the parameters for IP addresses and port numbers are correct
    if(robot_id == 0)
	{
        log_msg(ERR, LL_FATAL, "Invalid robot id (specify MAC address in hex, eg: 11a0f89fbec3).\n");
        usage(argv[0]);
    }

    if(gpsd_addr == NULL || gpsd_port == NULL)
	{
        log_msg(ERR, LL_FATAL, "Invalid IP address or port number specified for GPSD server.\n");
        usage(argv[0]);
    }

    log_msg(DBG, LL_DEBUG, "Opening serial port to access ADC on %s...\n", adc_serial);
    adc_fd = open(adc_serial, O_RDWR); //opens serial port to interface ADC device
    if(adc_fd == -1)
	{
        log_err(ERR, LL_FATAL, "Unable to open serial port to access ADC (%s).\n", adc_serial);
        exit(2);
    }

    // initialize connection to GPS daemon
    log_msg(DBG, LL_DEBUG, "Connecting to GPSD server at %s:%s...\n", gpsd_addr, gpsd_port);
    gpsdata = gps_open(gpsd_addr, gpsd_port);
    if(!gpsdata)
	{
        log_err(ERR, LL_FATAL, "Unable to connect to GPSD server (%s:%s).\n", gpsd_addr, gpsd_port);
        exit(2);
    }
    // register hook to position updates handler and init watcher mode
    gps_set_raw_hook(gpsdata, update_position);
    gps_query(gpsdata, GPS_INIT_STRING "\n");
    gps_poll(gpsdata);

    // populate unchangeable packet fields
    bzero(&packet, sizeof(packet));
    packet.status = PKT_STAT_INIT;
    get_mac_address(robot_id, packet.robot_id, sizeof(packet.robot_id));
    log_msg(DBG, LL_TRACE, "Robot id (hex): ");
    log_hex(DBG, LL_TRACE, packet.robot_id, sizeof(packet.robot_id));

    // header for logging of sensor data
    log_msg(DAT, LL_INFO, "Timestamp\tSensor\t[ADC]\tValue\t  GPS\tLatitude\t  Longitude\n");

    // main loop: process sensor samples, GPS position and send update messages
    for(;;)
	{
		// TODO: introduce more delay to ensure CO sensor gets enough recovery.------------------
		// 1-2 seconds should be good.
		// #define SLEEP_USEC          (2 * 1000000)
        usleep(SLEEP_USEC); //suspends execution for microsecond intervals

		// TODO: set system LEDs by sending some characters indicating various states ------------
		// GPS, USB, etc...

        // loops through ADC channels and sends a packet for each defined sensor
        for(i = 0; i < ADC_CHANNELS; i++)
		{
            // verify if this ADC channel is defined and usable (i.e. has a sensor attached)
            if(!sensor_defs[i].enabled ||
                sensor_defs[i].sensor_id == NULL ||
                sensor_defs[i].adc_handler == NULL)
            {
				//printf("[i=%u]", i);
                continue;
	    	}

            // gets Unix system timestamp (in seconds)
            packet.timestamp = time(NULL);

            // gets sensor id string from definition table
            bcopy(sensor_defs[i].sensor_id, packet.sensor_id, sizeof(packet.sensor_id));

            // reads ADC channel
            log_msg(DBG, LL_TRACE, "Reading sensor on ADC channel %d...\n", i);
            if(read_sensor(adc_fd, i, &adc_val) == -1)
			{
                log_msg(ERR, LL_WARN, "Failed to read sensor on ADC channel %d.\n", i);
                // if no values are available on the ADC channel(serial port) then
                // starts the repetition over again
                continue;
            }

            // computes real sensor value from ADC's discrete output
            packet.sensor_val.f = (*sensor_defs[i].adc_handler)(adc_val);
            log_msg(DBG, LL_TRACE, "Converted ADC#%d reading %u to real value %.2f.\n",
                    i, adc_val, packet.sensor_val.f);

            // execute the processing handler, if one is defined
            // (TODO: this feature can be used to trigger an action if value out-of-range)
            if(sensor_defs[i].value_handler != NULL)
			{
                if(sensor_proc = (*sensor_defs[i].value_handler)(packet.sensor_val.f))
                {
                    log_msg(DBG, LL_DEBUG, "Sensor value out-of-range (channel=%d, value=%f).\n",
                            i, packet.sensor_val.f);
				}
            }
	    	else
	    	{
                sensor_proc = 0;
            }

            // poll GPS daemon to retrieve geographical position
            log_msg(DBG, LL_TRACE, "Polling GPSD server...\n");
            if( gps_poll(gpsdata) == -1 ) // No GPS available
			{
                log_msg(ERR, LL_WARN, "Failed to query GPS...\n");
                //continue;

                // log data for current local sensor reading (data not sent)
                log_msg(DAT, LL_INFO, "%u\t%s\t0x%04x\t%.2f\t  N/A\tN/A\t  N/A\n",
                        packet.timestamp,
                        sensor_defs[i].sensor_id, adc_val, packet.sensor_val.f);

            }
            else	// GPS position is available
            {
				
				// determine whether GPS has usable fix or not -- see gpsd(8)
                if(gpsdata->fix.mode >= MODE_2D) SET_BIT(packet.status, PKT_STAT_GPS_FIX);
                else CLR_BIT(packet.status, PKT_STAT_GPS_FIX);

                packet.latitude.f = gpsdata->fix.latitude;
                packet.longitude.f = gpsdata->fix.longitude;

            	// log data for current local sensor reading, No GPS
            	log_msg(DAT, LL_INFO, "%u\t%s\t0x%04x\t%.2f\t  %s\t%.6f\t  %.6f\n",
                    packet.timestamp,
                    sensor_defs[i].sensor_id, adc_val, packet.sensor_val.f,
                    (gpsdata->fix.mode >= MODE_2D) ? "yes" : "no",
                    packet.latitude.f, packet.longitude.f);
			}

        } // sensor ADC channels loop

    } // main loop

    return 0;
} // end main(...)


